#! /usr/bin/python3

from .cmd_reader import reader
from .service import Service
from .builder import ZoneBuilder
from .process import launch_background_process, stop_background_process
from .conf.keybinding_service import KeybindingService

ZONES = "zones"
KEYFINDER = "keyfinder"


def main():
    cmd, args, kwargs = reader()
    if not any([cmd, args, kwargs]):
        service = Service()
        service.listen()

    if cmd == "config":
        _config_menu(*args, **kwargs)

    elif cmd == "start":
        launch_background_process(*args, **kwargs)

    elif cmd == "stop":
        stop_background_process()


def _config_menu(*args, **kwargs):
    if ZONES in args:
        zb = ZoneBuilder(*args, **kwargs)
        zb.main()

    elif KEYFINDER in args:
        print("Press Ctrl+C to quit.")
        kf = KeybindingService()
        kf.listen()


if __name__ == "__main__":
    main()



import os.path

HERE = os.path.abspath(os.path.dirname(__file__))


class CacheProperty:
    def __init__(self, fget) -> None:
        self.fget = fget

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = self.fget(obj)
        setattr(obj, self.fget.__name__, value)
        return value


class HelpMenu:
    def __init__(self) -> None:
        with open(os.path.join(HERE, "help_text.md"), "r") as f:
            self._raw_menu = f.read()

    def get(self, cmd):
        pass



import sys

from .help_menu import HelpMenu


def simple_isnan(item):
    try:
        float(item)
        return False
    except ValueError:
        return True


def _get_help(cmd=""):
    help_menu = HelpMenu()
    help_menu.get(cmd)
    sys.exit()


def reader():
    _input = sys.argv
    if len(_input) == 1:
        return "", [], {}

    cmd = _input[1:2].pop()

    args = _input[2::]
    kwargs = {}

    if cmd in ("-h", "--help"):
        return _get_help()
    elif any(h in args for h in ("-h", "--help")):
        return _get_help(cmd)

    kwarg_keys = reversed(
        [
            index
            for index, item in enumerate(args)
            if (item.startswith("-") and simple_isnan(item))
        ]
    )
    for k in kwarg_keys:
        try:
            value = args.pop(k + 1)
        except IndexError:
            value = True
        key = args.pop(k)
        kwargs[key] = value

    return cmd, args, kwargs



import os
import signal
import sys

from .service import Service
from .conf.settings import SETTINGS


def _check_pid(pid):
    if not pid:
        return False
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True


def launch_background_process(*args, **kwargs):
    # if there is a process already running, kill it
    _pid = SETTINGS.pid
    if _check_pid(_pid):
        print("background Snappy Zones process found: killing process...")
        stop_background_process(_pid)

    print("starting...")
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)
    except OSError as e:
        print(sys.stderr, f"FORK FAILED: {e.errno} ({e.strerror})")
        sys.exit(1)

    pid = str(os.getpid())
    SETTINGS.pid = pid
    print(f"started PID: {pid}")
    service = Service()
    service.listen()


def stop_background_process(pid=None):
    if not pid:
        pid = SETTINGS.pid
    if _check_pid(pid):
        print(f"stopping...")
        os.kill(pid, signal.SIGTERM)
        print(f"stopped PID: {pid}")
        os.remove(SETTINGS._pid_file)
    else:
        print("No background process was found. " "no action will be taken.")



from Xlib import X, XK
from Xlib.error import BadDrawable, XError
from Xlib.display import Display


def active_window(display, window_id=None):
    if not window_id:
        window_id = (
            display.screen()
            .root.get_full_property(
                display.intern_atom("_NET_ACTIVE_WINDOW"), X.AnyPropertyType
            )
            .value[0]
        )
    try:
        return display.create_resource_object("window", window_id)
    except XError:
        return None


def geometry_deltas(window):
    """The window of an app usually sits within an Xorg parent frame, and we
    want to fit that parent frame to the zone and not the inner window (so
    decorations like borders are properly handled). I can't seem to get that
    window to update directly, so we'll update the child window using the
    difference b/t the parent and child so to fit the final result
    correctly."""
    wg = window.get_geometry()
    dx, dy, dw, dh = 0, 0, 0, 0
    parent = window.query_tree().parent
    pg = parent.get_geometry()
    dx = pg.x - wg.x
    dy = pg.y - wg.y
    dw = pg.width - wg.width
    dh = pg.height - wg.height
    return dx, dy, dw, dh


def shift_window(self, keysym, stretch=False):
    try:
        display = Display()
        zone_profile = self.zp
        window = active_window(display)
        wg = window.get_geometry()
        dx, dy, dw, dh = geometry_deltas(window)
        pg = window.query_tree().parent.query_tree().parent.get_geometry()
        zone = zone_profile.find_zone(pg.x + pg.width / 2, pg.y + pg.height / 2, keysym)
        if window and zone:
            window.configure(
                x=zone.x,
                y=zone.y,
                width=zone.width - dx,
                height=zone.height - dy,
                stack_mode=X.Above,
            )
            display.sync()
    except BadDrawable:
        pass


def snap_window(self, x, y):
    try:
        display = Display()
        zone_profile = self.zp
        window = active_window(display)
        dx, dy, dw, dh = geometry_deltas(window)
        zone = zone_profile.find_zones(self, x, y)
        if window and zone:
            window.configure(
                x=zone.x,
                y=zone.y,
                width=zone.width - dx,
                height=zone.height - dy,
                stack_mode=X.Above,
            )
            display.sync()
    except BadDrawable:
        pass



import json
import os
import sys
import threading

from Xlib import X
from Xlib.display import Display
from Xlib.error import ConnectionClosedError

from .process import _check_pid, launch_background_process
from .snap import active_window
from .conf.settings import SETTINGS

HERE = os.path.abspath(os.path.dirname(__file__))


def _get_recursive(obj, args, default=None):
    """Apply successive .get() calls to container obj and return result if
    something is found, else return specified default value"""
    if not args:
        return obj
    if isinstance(obj, (dict, list, tuple)):
        try:
            sub_obj = obj.__getitem__(args[0])
            return _get_recursive(sub_obj, args[1:], default)
        except (KeyError, IndexError):
            return default
    return default


class ZoneBuilder:
    def __init__(self, *args, **kwargs) -> None:
        self.zones = []
        self.display = Display()
        self.screen = self.display.screen()

        if count := kwargs.get("-n"):
            self.add(count)

    def _main_action(self):
        return input("specify action:\n$ ")

    def _get_count(self):
        return input("specify count:\n$ ")

    def main(self):
        action = self._main_action()

        if action == "save":
            self.save

        elif action == "exit":
            self.terminate()

        else:
            print("action not understood. Please try again.\n")

        self.main()

    @property
    def save(self):
        results = []
        for zone in self.zones:
            # for some reason this works better with
            # a new display object on each iteration
            window = active_window(Display(), zone.id)
            pg = window.query_tree().parent.query_tree().parent.get_geometry()
            results.append(
                {"x": pg.x, "y": pg.y, "width": pg.width, "height": pg.height}
            )
        SETTINGS.zones = results

        # if service is running, restart with new zones
        if _check_pid(SETTINGS.pid):
            print("restarting background process for updated zone settings...")
            launch_background_process()

        self.terminate()

    def add(self, count):
        previous_settings = SETTINGS.zones  # list of dicts
        for i in range(int(count)):
            x = _get_recursive(previous_settings, (i, "x"), default=10)
            y = _get_recursive(previous_settings, (i, "y"), default=10)
            width = _get_recursive(previous_settings, (i, "width"), default=500)
            height = _get_recursive(previous_settings, (i, "height"), default=250)
            window = self.screen.root.create_window(
                10,
                10,
                500,
                250,
                1,
                self.screen.root_depth,
                background_pixel=750000,
                event_mask=X.ExposureMask | X.KeyPressMask,
            )
            window.map()
            window.configure(x=x, y=y, width=width, height=height, stack_mode=X.Above)
            self.display.sync()
            self.zones.append(window)
        thread = threading.Thread(target=self.loop)
        thread.daemon = True
        thread.start()

    def loop(self):
        while True:
            try:
                e = self.display.next_event()
            except ConnectionClosedError:
                # raised when X button is pressed
                #
                # TODO: figure out way to allow exit of single window without
                # having to kill the entire application (currently get a
                # "socket_error" if you try to add a window after exiting
                # an already made window)
                #
                # for now, just kill everything
                print()
                os._exit(1)

    def terminate(self):
        sys.exit()



from Xlib import XK

from .conf.settings import SETTINGS

MEAN_PIXEL_TOLERANCE = 10


def mean(lst):
    return sum(lst) / len(lst)


class Zone:
    def __init__(self, x, y, width, height) -> None:
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def check(self, x, y):
        if (self.x <= x <= self.x + self.width) and (
            self.y <= y <= self.y + self.height
        ):
            return True
        return False

    @property
    def corners(self):
        return [
            (self.x, self.y),
            (self.x + self.width, self.y),
            (self.x + self.width, self.y + self.height),
            (self.x, self.y + self.height),
        ]


class ZoneProfile:
    def __init__(self, zones) -> None:
        self.zones = zones

    def find_zones(self, service, x, y):
        _zones = []
        for coordinate in service.coordinates:
            for item in self.zones:
                if item.check(*coordinate) and item not in _zones:
                    _zones.append(item)
                    break

        if not _zones:
            return None

        if len(_zones) == 1:
            return _zones.pop()

        x_min_zone = min((i for i in _zones), key=lambda i: i.x)
        y_min_zone = min((i for i in _zones), key=lambda i: i.y)

        # if all zones are in the same row
        if abs(mean(set(i.x for i in _zones)) - x_min_zone.x) < MEAN_PIXEL_TOLERANCE:
            width = x_min_zone.width
            height = sum(i.height for i in _zones)

        # if all zones are in the same column
        elif abs(mean(set(i.y for i in _zones)) - y_min_zone.y) < MEAN_PIXEL_TOLERANCE:
            width = sum(i.width for i in _zones)
            height = y_min_zone.height

        # stretch first zone into last zone
        elif len(_zones) == 2:
            initial_zone = self.find_zone(*service.coordinates[0])
            final_zone = self.find_zone(*service.coordinates[-1])
            slope = abs(
                (initial_zone.y - final_zone.y) / (initial_zone.x - final_zone.x)
            )
            if slope > 1:  # means we're stretching the height
                x = x_min_zone.x
                y = y_min_zone.y
                width = initial_zone.width
                height = initial_zone.height + final_zone.height
            else:  # means we're stretching the width
                x = x_min_zone.x
                y = initial_zone.y
                width = initial_zone.width + final_zone.width
                height = initial_zone.height
            return Zone(x, y, width, height)

        # return a zone which covers all zones
        else:
            width, height = 0, 0
            for z in _zones:
                if z.corners[1][0] - x_min_zone.x > width:
                    width = z.corners[1][0] - x_min_zone.x
                if z.corners[3][1] - y_min_zone.y > height:
                    height = z.corners[3][1] - y_min_zone.y
            return Zone(x_min_zone.x, y_min_zone.y, width, height)
        return Zone(x_min_zone.x, y_min_zone.y, width, height)

    def find_zone(self, x, y, shift=None):
        for index, item in enumerate(self.zones):
            if item.check(x, y):
                if not shift:
                    obj_i = index
                    return self._shift_and_return(obj_i)
                elif shift == XK.XK_Left:
                    obj_i = (index - 1) % len(self.zones)
                    return self._shift_and_return(obj_i)

                elif shift == XK.XK_Right:
                    obj_i = (index + 1) % len(self.zones)
                    return self._shift_and_return(obj_i)
        return None

    def _shift_and_return(self, obj_i):
        obj = self.zones[obj_i]
        self.zones = self.zones[obj_i:] + self.zones[:obj_i]
        return obj

    @staticmethod
    def from_file():
        if data := SETTINGS.zones:
            return ZoneProfile([Zone(**obj) for obj in data])
        return ZoneProfile([])



import os.path

from setuptools import setup, find_packages

HERE = os.path.dirname(os.path.abspath(__file__))

def read_requirements(requirements_file_path):
    """Return dependencies from a requirements file as a list.

    Read a requirements '.txt' file, where dependencies are separated by a new line.
    Removes all comments and options for pip, and return as a list of dependencies.

    :return:    requirements
    :rtype:     list
    """
    with open(requirements_file_path, "r") as f:
        data = f.readlines()
    data = [i[: i.find("#")] if "#" in i else i for i in data]
    data = [i.strip() for i in data if i.strip()]
    data = [i for i in data if not i.startswith("-")]
    return data


with open("README.md", "r") as f:
    long_description = f.read()

setup(
    author="Michael Green",
    author_email="1mikegrn@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    description="a Linux window manager based on Microsoft's FancyZones",
    entry_points={"console_scripts": ["snappy=snappyzones.__main__:main"]},
    include_package_data=True,
    install_requires=read_requirements(os.path.join(HERE, "requirements.txt")),
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="snappyzones",
    packages=find_packages(where="src"),
    package_data={"": ["*.json", "*.txt"]},
    package_dir={"": "src"},
    python_requires=">=3.8",
    tests_require=read_requirements(os.path.join(HERE, "requirements_testing.txt")),
    url="https://github.com/1mikegrn/snappyzones",
    version="0.0.6",
)



from Xlib import X, XK
from Xlib.ext import record
from Xlib.display import Display
from Xlib.protocol import rq

from .snap import snap_window, shift_window
from .zoning import ZoneProfile
from .conf.settings import SETTINGS


class Service:
    def __init__(self) -> None:
        self.active_keys = {
            XK.string_to_keysym(key): False for key in SETTINGS.keybindings
        }
        self.modifying_keys = {
            XK.string_to_keysym(key): False for key in SETTINGS.keybinding_modifier
        }

        self.zp = ZoneProfile.from_file()
        self.coordinates = Coordinates()

        self.display = Display()
        self.root = self.display.screen().root

        self.context = self.display.record_create_context(
            0,
            [record.AllClients],
            [
                {
                    "core_requests": (0, 0),
                    "core_replies": (0, 0),
                    "ext_requests": (0, 0, 0, 0),
                    "ext_replies": (0, 0, 0, 0),
                    "delivered_events": (0, 0),
                    "device_events": (X.KeyReleaseMask, X.ButtonReleaseMask),
                    "errors": (0, 0),
                    "client_started": False,
                    "client_died": False,
                }
            ],
        )

        self.display.record_enable_context(self.context, self.handler)
        self.display.record_free_context(self.context)

    def handler(self, reply):
        data = reply.data
        while len(data):

            event, data = rq.EventField(None).parse_binary_value(
                data, self.display.display, None, None
            )

            if event.type in (X.KeyPress, X.KeyRelease):
                keysym = self.display.keycode_to_keysym(event.detail, 0)
                if keysym in self.active_keys:
                    self.active_keys[keysym] = (
                        True if event.type == X.KeyPress else False
                    )
                elif keysym in self.modifying_keys:
                    self.modifying_keys[keysym] = (
                        True if event.type == X.KeyPress else False
                    )

            if all(self.active_keys.values()):
                self.coordinates.add(event.root_x, event.root_y)
                if (event.type, event.detail) == (X.ButtonRelease, X.Button1):
                    snap_window(self, event.root_x, event.root_y)
                elif event.type == X.KeyPress:
                    keysym = self.display.keycode_to_keysym(event.detail, 0)
                    shift_window(self, keysym)

            else:
                self.coordinates.clear()

    def listen(self):
        while True:
            self.root.display.next_event()


class Coordinates:
    def __init__(self) -> None:
        self.x = []
        self.y = []

    def __getitem__(self, item):
        """returns an (x,y) coordinate"""
        return self.x[item], self.y[item]

    def __iter__(self):
        """iterate over (x,y) coordinates"""
        return zip(self.x, self.y)

    def add(self, x, y):
        self.x.append(x)
        self.y.append(y)

    def clear(self):
        self.x = []
        self.y = []



import os.path
import json


HERE = os.path.abspath(os.path.dirname(__file__))


class Settings:
    @property
    def pid(self):
        if res := self._read(self._pid_file):
            return int(res)
        return None

    @pid.setter
    def pid(self, _pid):
        self._write(self._pid_file, str(_pid))

    @property
    def zones(self):
        if _zones := self._read(self._zone_file):
            return json.loads(_zones)
        return None

    @zones.setter
    def zones(self, _zones):
        self._write(self._zone_file, json.dumps(_zones, indent=2, sort_keys=True))

    @property
    def keybindings(self):
        _data = self._read(self._keybindings_file)
        results = [
            char
            for char in _data.split("\n")
            if not char.startswith(("#", "!")) and char != ""
        ]
        if results:
            return results
        return ["Shift_L"]  # a default in case someone tries to be funny

    @property
    def keybinding_modifier(self):
        _data = self._read(self._keybindings_file)
        results = [
            char.replace("! ") for char in _data.split("\n") if char.startswith("!")
        ]
        if results:
            return results
        return ["Control_L"]  # a default in case someone tries to be funny

    @property
    def _raw_keybindings(self):
        _data = self._read(self._keybindings_file)[1:]
        return [char.replace("# ", "") for char in _data.split("\n")]

    @property
    def _keybindings_file(self):
        return os.path.join(HERE, "keybindings.txt")

    @property
    def _zone_file(self):
        return os.path.join(HERE, "zones.json")

    @property
    def _pid_file(self):
        return os.path.join(HERE, ".pid")

    def _read(self, path, default=None):
        if os.path.isfile(path):
            with open(path, "r") as f:
                return f.read()
        return default

    def _write(self, path, obj):
        with open(path, "w") as f:
            f.write(obj)


SETTINGS = Settings()



import os

from Xlib import X, XK
from Xlib.ext import record
from Xlib.display import Display
from Xlib.protocol import rq

from .settings import SETTINGS


def set_keybindings(*args, **kwargs):
    SETTINGS.keybindings = args


class KeybindingService:
    def __init__(self) -> None:
        self.display = Display()
        self.root = self.display.screen().root

        self.staging = ""
        self.keys = []

        self.keymap = {
            XK.string_to_keysym(key): key for key in SETTINGS._raw_keybindings
        }

        self.context = self.display.record_create_context(
            0,
            [record.AllClients],
            [
                {
                    "core_requests": (0, 0),
                    "core_replies": (0, 0),
                    "ext_requests": (0, 0, 0, 0),
                    "ext_replies": (0, 0, 0, 0),
                    "delivered_events": (0, 0),
                    "device_events": (X.KeyReleaseMask, X.ButtonReleaseMask),
                    "errors": (0, 0),
                    "client_started": False,
                    "client_died": False,
                }
            ],
        )

        self.display.record_enable_context(self.context, self.handler)
        self.display.record_free_context(self.context)

    def handler(self, reply) -> None:
        data = reply.data
        while len(data):
            event, data = rq.EventField(None).parse_binary_value(
                data, self.display.display, None, None
            )

            if event.type == X.KeyPress:
                keycode = event.detail
                keysym = self.display.keycode_to_keysym(keycode, 0)
                if keysym in self.keymap and self.keymap[keysym] not in self.keys:
                    self.keys.append(self.keymap[keysym])
                end_str = f"\rKEYBINDING: {' + '.join(self.keys)}   LAST_KEY: "
                clear_str = "\r" + (" " * int(os.get_terminal_size().columns * 0.75))
                print(clear_str, end=end_str)

            elif event.type == X.KeyRelease:
                keycode = event.detail
                keysym = self.display.keycode_to_keysym(keycode, 0)
                if keysym in self.keymap and self.keymap[keysym] in self.keys:
                    self.keys.remove(self.keymap[keysym])
                end_str = f"\rKEYBINDING: {' + '.join(self.keys)}   LAST_KEY: "
                clear_str = "\r" + (" " * int(os.get_terminal_size().columns * 0.75))
                print(clear_str, end=end_str)

            else:
                print(end="\r")

    def listen(self) -> None:
        while True:
            self.display.next_event()
